"""
This is a database of unique files which may exist in different locations.
In addtiion the file entities can be segmented to fit onto multiple discs.
"""
import datetime as dt
import json
from pathlib import Path, PurePosixPath
from os import fsdecode, fsencode, getcwd, lstat, readlink, stat_result, getcwd
import uuid

try:
    from scandir import walk
except ImportError:
    from os import walk

from .consts import *
from .abstract_file_db import AbstractFileDatabase
from .file_db import FileDatabase
from .file_entry import FileEntryType, FileEntry
from .hash_file_entry import HashFileEntries, HashFileEntry


class HashDatabase(AbstractFileDatabase):
    """This consists of making a database which has as a key each hash entry with multiple
    entries for each file.
    In addition it handles segmented the database for conversion to a set of ISO files"""

    def __init__(self, file_db: FileDatabase, iso_path_root):
        self.iso_path_root = iso_path_root
        self.db_path = Path(DB_FILENAME)
        self.version = DATABASE_VERSION
        self.segment_size = None  # DB is started not segmented
        self.last_disc_number = None  # This starts as a non segmented archive
        # segmented or not is None or not
        try:
            self.entries = HashFileEntries.create(self.iso_path_root, file_db.path)
            self.update(file_db)
        except AttributeError:
            self.entries = HashFileEntries.create(self.iso_path_root, None)

    def save(self, catalogue_name=DB_FILENAME):
        """Save the current catalogue to file as a JSON file.
        It should be possible to reread this file later and recreate this record and a complete archive."""
        self.guid = uuid.uuid4()  # a second save will have a different guid as the structure is mutable and this
        # ensures that each saved file is uniquely identifiable.
        filename = Path(getcwd()) / catalogue_name
        data = {
            # Todo If use original path then need to create dynamically the path to match
            #"OriginalPath": str(self.path),
            "date": str(dt.datetime.utcnow().isoformat()),
            "guid": str(self.guid),
            "version": self.version,
            "files": json.loads(self.entries.to_json()),
            # List of directories are derived from file paths
            # List of catalogues as GUID, the first will be this catalogue
        }
        with filename.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, sort_keys=True, indent=4)

    def segment(self, size, catalogue_size):
        """
        For a catalogue will place each file onto a disc.
        This will overwrite the segments if carrie out repeatedly.
        :param size:
        :return:
        """
        # Deal with differing types of segment size
        self.str_segment_size = size
        new_size = interpret_disc_capacity(size)
        self.int_segment_size = new_size
        self.segment_size = new_size
        # Store the size of the catalogue
        self.catalogue_size = catalogue_size
        OVERHEAD = (
            500000 + catalogue_size
        )  # This is the number of bytes of overhead that will be used for size of iso file
        # TODO needs a method to deal more accruately with directory sizes
        FILE_OVERHEAD = (
            2048
        )  # This is the number of bytes of overhead this is allocated for each file in addition
        # to the actual contents of the file.  Needs to cover directory entries.
        min_size = OVERHEAD + FILE_OVERHEAD + next(iter(self.files())).size
        if new_size <= min_size:
            raise odarchiveError(f"Disc too small {new_size:,}, cannot fit first file on with overhead {min_size:,}.")

        count = OVERHEAD # Count the number of bytes used
        self.last_disc_number = 0
        for entry in self.files():
            size_on_disc = (
                (2048 + entry.size) // 2048
            ) * 2048 + FILE_OVERHEAD
            if (count + size_on_disc) >= self.segment_size:
                # If too big to fit file in segment then start a new disc
                count = OVERHEAD
                self.last_disc_number += 1
            # Only add
            entry.disc_num = self.last_disc_number
            count += size_on_disc
            if count > self.segment_size:
                # if file is too big to fit on a single disc with overhad
                raise odarchiveError(f"Disc too small {new_size:,}, cannot fit file {entry.filename} on disc {self.last_disc_number} with overhead {count:,}.")

    @property
    def is_segmented(self):
        return self.last_disc_number is not None

    def update(self, file_db):
        """
        Iterates a file entry database to get each entry and add to hash database
        """
        for entry in file_db.files():
            self.entries.add_hash_file(entry)

    def files(self, disc_num=None):
        """Extend class with a disc number segemtn"""
        for entry in self.entries.values():
            if disc_num is None:  # without a disc num specification return all files
                yield entry
            elif entry.disc_num == disc_num:  # only get entries for this disc_num
                yield entry

    def get_info(self, for_disc_num = None):
        """Returns summary information on an archive. Uses introspection"""
        count_files = 0
        size_files = 0
        disc_nums = set()
        entries_no_disc = 0
        largest_file = 0
        max_dir_length = 0
        longest_dir = ""
        dirs = set()
        for entry in self.files():
            count_files += 1
            size_files += entry.size
            if for_disc_num is None or for_disc_num == entry.disc_num:
                count_files += 1
                size_files += entry.size
                if self.is_segmented:
                    if entry.disc_num is None:
                        entries_no_disc += 1
                    else:
                        disc_nums |= {entry.disc_num}
                if entry.size > largest_file:
                    largest_file = entry.size
                # Each entry may have multiple directory entries
                for this_file in entry.filenames:
                    this_dir = PurePosixPath(this_file).parent
                    dirs = dirs | {this_dir}
                    length = len(Path(this_dir).parts) - 2
                    if length > max_dir_length:
                        longest_dir = this_dir.parent
                        max_dir_length = length
        # Format answer
        result = super().get_info()
        result += f"Data size       = {size_files:,} bytes\n"
        result += f"Is segmented    = {self.is_segmented}\n"
        if for_disc_num is None:
             result += f'>>>>>>>>> For all files in all discs <<<<<<<<<<<<<<<<\n'
        else:
            if self.is_segmented:  # Add information about disc size
                result += f'>>>>>>>>> Disc {for_disc_num} <<<<<<<<<<<<<<<<\n'
            else:
                result += f'>>>>>>>>> Error asking for disc {for_disc_num} but disc not segmented  <<<<<<<<<<<<<<<<\n'
        if self.is_segmented:  # Add information about disc size
            if isinstance(self.str_segment_size, str):  # Then can interpret text
                result += (
                    f"  Disc segment size = {self.str_segment_size}, {self.int_segment_size:,} bytes\n"
                )
            else:  # Can only write size
                result += f"  Disc segment size = {self.int_segment_size:,} bytes\n"
            result += f"  Catalogue size = {self.catalogue_size:,} bytes\n"
            result += f"  Number of discs = {self.last_disc_number+1:,}\n"
        if entries_no_disc > 0 and self.is_segmented:
            result += (
                f"  ERROR: {entries_no_disc} entries have not been allocated a disc number and should have been.\n"
            )
        result += f"Number of files = {count_files:,}\n"
        result += f"  Largest file  = {largest_file:,}\n"
        result += f"Number of dirs  = {len(dirs)}\n"
        result += f"Max dir depth   = {max_dir_length} (on source file system)\n"
        result += f" Dir =: {longest_dir}\n"
        if hasattr(self,'guid'):
            result += (
                f"guid = {self.guid}\n"
            )
        result += f"Database Version = {self.version}\n"
        return result



import datetime as dt
from collections import OrderedDict
from enum import Enum
import json
from mmap import mmap, ACCESS_READ
from pathlib import Path, PurePosixPath
from os import fsdecode, fsencode, getcwd, lstat, readlink, stat_result
from os.path import normpath
import time

import dill
from stat import S_ISLNK, S_ISREG

from .consts import *


from .tools import mangle_file_for_iso9660, mangle_dir_for_iso9660


class HashFileEntries(OrderedDict):
    """This is a collection of HashFileEntries
    In fact you can only create a new HashFileEntry with reference to a collection
    """
    @classmethod
    def create(cls, iso_path_root, path):
        """ Did this to get around issue with loading pickled object that is derived from an OrderedDict"""
        result = cls()
        result.iso_path_root = iso_path_root
        result.path = path
        return result

    @classmethod
    def create_from_json(cls, iso_path_root, files_in_db, parent):
        """ Reading in entries from json"""
        result = cls()
        result.iso_path_root = iso_path_root
        result.path = ''  # TODO Preserve path
        print(files_in_db)
        for hash, entry in files_in_db.items():
            # Sort out is_sgemented and last_disc_number in parent object
            try:
                disc_num = int(entry['disc_num'])
                if parent.last_disc_number is None:
                    parent.last_disc_number = disc_num
                else:
                    if disc_num > parent.last_disc_number:
                        parent.last_disc_number = disc_num
            except KeyError:
                disc_num = None
            filenames = []
            for filename in entry['filenames']:
                filenames.append(filename)
            result[hash] = HashFileEntry(
                result,
                hash,
                filenames,
                entry['size'],
                entry['mtime'],
                0,#         catalogue_num=catalogue_num,
                disc_num #         disc_num=this_entry.disc_num,
            )
            #Add extra filenames
        return result

    def entry_to_path(self, this_entry):
        """ Converts a fileEntry object to an ISO path via relative path

        To reverse use orig_root / self.relative_path
        """
        return self.iso_path_root / this_entry.relative_path

    def add_hash_file(self, this_entry, disc_num=None, catalogue_num=None):
        """If the hash file already exists then this will not create a new entry but simply
        append this file to the HashFileEntry.

        this_file is a FileEntry
        """
        #  Does it already exist
        try:
            existing_entry = self[this_entry.file_hash]  # Look up by File hash
            if existing_entry.has_file_path(
                self.entry_to_path(this_entry)
            ):  # Already is entered then ignore otherwise add to list of directories
                pass
            else:
                existing_entry.add_path(self.entry_to_path(this_entry))
        except AttributeError:  # file_hash is not yet defined
            raise odarchiveError(
                f"Adding {self.entry_to_path(this_entry)} with no file_hash - run calculate_file_hash()"
            )
        except KeyError:  # Adding a new entry to the Dictionary
            if this_entry.file_hash:
                self[this_entry.file_hash] = HashFileEntry(
                    self,
                    this_entry.file_hash,
                    self.entry_to_path(this_entry),
                    size=this_entry.size,
                    mtime=this_entry.mtime,
                    catalogue_num=catalogue_num,
                    disc_num=this_entry.disc_num,
                )
            else:  # Adding a file without a hash must be an error
                raise odarchiveError(
                    f"Adding {self.entry_to_path(this_entry)} with no Hash - probable programming error"
                )

    def AddHashFileInAnotherCatalogue(
        self, file_hash, filename, size=None, mtime=None, file_type=None, disc_num=None
    ):
        """If the hash file already exists then this will change the entry to note that it is stored
        in a prior catalogue.
        """
        pass

    def sort(self):
        """Sort dictionary so that when it is written out by a json method it all works"""
        # # dictionary sorted by length of the key string
        # >> > OrderedDict(sorted(d.items(), key=lambda t: len(t[0])))
        # OrderedDict([('pear', 1), ('apple', 4), ('orange', 2), ('banana', 3)])
        pass

    def to_json(self):
        header = "{\n"
        result = ""
        footer = "}\n"
        for i, entry in enumerate(self.values()):
            if i:
                result += " , "  # Can;t have trailing comma
            result += entry.to_json_entry() + "\n"
        return header + result + footer

    def dir_entries(self, disc_num=None):
        """Iterate over all files in the hash entry list and build a dictionary of all of them"""
        result = {}

        def update_dir_list(udf_path):
            nonlocal result
            if str(udf_path) not in result:
                # Make sure parent is in
                if len(udf_path.parts) > 2:  # Has parent
                    update_dir_list(
                        udf_path.parent
                    )  # Recursive parent traversal to make sure every
                    # level of directory is included.
                result[str(udf_path)] = ""

        for entry in self.values():
            if disc_num is None or disc_num == entry.disc_num:
                update_dir_list(
                    entry.udf_absolute_path.parent
                )  # Only add parent but do it recursively
            # else Do not add this directory to this entry
        return result


class HashFileEntry:
    """This represents a single duplicated file.  In can either be in this catalogue or
    in another catalogue.  You cannot create an entry without know the hash of the file.

    The filenames of each entry and duplicate are stored in a filenames dictionary.
    filename can either be a single element or a list of filenames
    """

    def __init__(
        self,
        parent,
        file_hash,
        filename,
        size=None,
        mtime=None,
        disc_num=None,
        catalogue_num=None,
    ):
        # In memory, "filename" should be a relative UDF Path
        self.parent = parent  # eg a HashFileEntries
        if type(filename) is list:
            self.filenames = {}
            for this_file in filename:
                self.filenames[this_file] = {}
        else:
            self.filenames = {
                str(filename): {}
            }  # a UDF absolute path which is stored on a blank catalogue
        self.size = size
        self.mtime = mtime
        self.file_hash = file_hash
        self._disc_num = None
        self.disc_num = disc_num
        self.catalogue_num = (
            catalogue_num
        )  # If None or 0 then in this catalogue otherwise in another catalogue
        #  You will need to look up the catalogue number to the GUID of the cataloge at the start of the catalogue

    @property
    def filename(self):
        """This represents the filename on disc of the hash file.  There may be many filenames eg copies, links
        but only one will be stored on disc"""
        return PurePosixPath(next(iter(self.filenames.keys())))

    @property
    def disc_num(self):
        if hasattr(self, "_disc_num"):
            return self._disc_num
        else:
            return None

    @disc_num.setter
    def disc_num(self, disc_num):
        self._disc_num = disc_num  # Rely on Archive level lock for overwriting

    def __str__(self):
        return f"{self.filename}, {self.hash}"

    @property
    def file_system_path(self):
        """Returns native files system absolute path"""
        return PurePosixPath(self.parent.path) / self.relative_filename

    @property
    def udf_absolute_path(self):
        """Given a UDF media (which might be 1..n) this then returns the absolute path on that media"""
        return self.parent.iso_path_root / self.filename

    @property
    def relative_filename(self):
        """returns the data part of the path without /DATA prefix"""
        return PurePosixPath(self.filename.relative_to(self.parent.iso_path_root))

    @property
    def iso9660_path(self):
        # Assume PlainBuild UDF ie written in one pass not incrementally ie suffix is ";1"
        # TODO do folding so that if the iso9660_path has been already used another is used.
        # Needs to be done through parent as the parent controls the restart of reusage cache
        file_start = self.parent.iso_path_root / self.filename
        mangled_dirs = [
            mangle_dir_for_iso9660(part, 3) for part in file_start.parent.parts
        ]
        mangled_dirs[0] = mangled_dirs[0][1:]  # Get rid of leading underscore
        mangled = (
            mangle_file_for_iso9660(str(file_start.stem), 1)[0]
            + "."
            + mangle_file_for_iso9660(str(file_start.suffix)[1:], 1)[0]
        )

        return "/".join(mangled_dirs) + "/" + mangled + ";1"

    def to_json_entry(self):
        """Returns a string which is the entry in an object
        """
        filename_list = ""
        for i, filename in enumerate(self.filenames):
            if i:
                filename_list += ", "
            filename_list += f'"{filename}" : null'
        if self.disc_num is not None:  # 0 is a valid disc_num
            disc_num = f'    "disc_num" : {self.disc_num},\n'
        else:
            disc_num = ""
        return (
            f'"{self.file_hash}"'
            + ": {\n"
            + '    "filenames" : {\n'
            + filename_list
            + "    },"
            + disc_num
            + f'    "size" : {self.size},\n'
            + f'    "mtime" : "{dt.datetime.fromtimestamp(self.mtime).strftime("%Y-%m-%dT%H:%M:%S")}"\n'
            + "}\n"
        )

    def add_path(self, this_path):
        self.filenames[str(this_path)] = {}

    def has_file_path(self, this_path):
        """A has file entry has multiple paths this tests if a UDF path has been stored."""
        # TODO should probably test UDF relative path
        try:
            a = self.filenames[str(this_path)]
            return True
        except KeyError:
            return False


def iso9660_dir(this_dir):
    """
    A function to convert a UDF full path string to an ISO 9660 V3 compliant version.

    Parameters:
     this_dir - string UDF file name
    Returns:
     The truncated and translated name suitable for the ISO interchange level
     specified.
    """
    mangled_dirs = [mangle_dir_for_iso9660(part, 3) for part in Path(this_dir).parts]
    mangled_dirs[0] = mangled_dirs[0][1:]  # Get rid of leading underscore
    result = "/".join(mangled_dirs)
    if result == "":
        result = "/"
    return result



import hashlib
import re

# 1: 'version' field added
# 2: entry 'file_type' field added; symlinks now treated correctly
DATABASE_VERSION = 1
DB_FILENAME = "catalogue.json"

HASH_FUNCTION = hashlib.sha512
# Mostly used for importing from saved hash files
EMPTY_FILE_HASH = (
    "cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce"
    "47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e"
)

SHA512_HASH_PATTERN = re.compile(r"^[0-9a-fA-F]{128}$")

HASH_FILENAME = "SHA512SUM"


class odarchiveError(Exception):
    pass

def interpret_disc_capacity(size):
    """
    Converts a size parameter into a number of bytes
    :param size: either text or an int representing a number of bytes
    :return:
    """

    def p(comma_num):
        return int(comma_num.replace(",", ""))

    size_as_text = str(size).lower()
    if size_as_text == "cd":
        return p("737,280,000")
    elif size_as_text == "dvd":  # DVD-R SL DVD+R is slightly bigger
        return p("4,707,319,808")
    elif size_as_text == "bluray" or size_as_text == "bd":
        # return p("25,025,314,816   ")  # Theoretical maximum but ISO's are create slightly bigger
        return p("23,000,000,000   ")  # Try < 99% 99% still too large
    elif size_as_text == "bd-dl":
        return p("50,050,629,632")
    elif size_as_text == "bd-xl":
        return p("100,103,356,416")
    elif size_as_text == "bd - xx":
        return p("128,001,769,472")
    else:
        return size



from enum import Enum
from mmap import mmap, ACCESS_READ
from pathlib import Path, PurePosixPath
from os import fsencode, lstat, readlink, stat_result

from stat import S_ISLNK, S_ISREG

from .consts import *


class FileEntryType(Enum):
    TYPE_FILE = 0
    TYPE_SYMLINK = 1


class FileEntry:
    """This represents each file stored.
    It is also meant to deal with:
        - both symlinks and files
        - Building a catalogue before you have the hashes for each file
    """

    def __init__(
        self, parent, filename, size=None, mtime=None, type=None, disc_num=None
    ):
        # In memory, "filename" should be an absolute Path
        self.parent = parent
        self.filename = filename
        self.size = size
        self.mtime = mtime
        self.type = type
        self._disc_num = None
        self.disc_num = disc_num

    @property
    def disc_num(self):
        if hasattr(self, "_disc_num"):
            return self._disc_num
        else:
            return None

    @disc_num.setter
    def disc_num(self, disc_num):
        if self.disc_num is None:
            self._disc_num = disc_num  # Can set it the first time

    def exists(self):
        return self.filename.is_file() or self.filename.is_symlink()

    def verify(self):
        current = FileEntry(self.parent, self.filename)
        return self == current

    def __eq__(self, other):
        """Weak would be stronger with hashes"""
        if isinstance(other, FileEntry):
            return (
                self.size == other.size
                and self.mtime == other.mtime
                and (self.type == other.type)
            )
        return super().__eq__(other)

    def update_attrs(self):
        s = lstat(str(self.filename))
        self.size, self.mtime = s.st_size, s.st_mtime

    def update_type(self):
        if self.filename.is_symlink():
            self.type = FileEntryType.TYPE_SYMLINK
        else:
            # Treat it as a file even if it's missing.
            self.type = FileEntryType.TYPE_FILE

    def update(self):
        self.update_attrs()
        self.update_type()

    def __str__(self):
        return str(self.filename)

    def __hash__(self):
        return hash(self.filename)

    @property
    def relative_path(self):
        """Returns relative path to parent directory"""
        return PurePosixPath(self.filename.relative_to(self.parent.path))

    def calculate_file_hash(self):
        if self.filename.is_file():
            if lstat(str(self.filename)).st_size > 0:
                with self.filename.open("rb") as f:
                    with mmap(f.fileno(), 0, access=ACCESS_READ) as m:
                        self.file_hash = HASH_FUNCTION(m).hexdigest()
            else:
                self.file_hash = EMPTY_FILE_HASH
        elif self.filename.is_symlink():
            # The link target will suffice as the "contents"
            target = readlink(str(self.filename))
            self.file_hash = HASH_FUNCTION(fsencode(target)).hexdigest()



"""This is a database of files.

It is centred on the file system.  Each file has a separate entry  It can be enhanced.
This is a temporary structure and although can be preserved by pickle is not intended to have a long duration.
So it does not need versioning.


Things it doesn't do:

- Segment the database.  See hash_db
"""
from pathlib import Path
from os import lstat
from scandir import walk
from sys import stderr

from .abstract_file_db import AbstractFileDatabase
from .file_entry import FileEntry

def do_hash(entry):
    """Make an easy parallel task"""
    entry.calculate_file_hash()

class FileDatabase(AbstractFileDatabase):

    def __init__(self, path: Path):
        super().__init__(path)
        self.path = path.absolute()

    def _find_changes(self):
        """
        Walks the filesystem. Identifies noteworthy files -- those
        that were added, removed, or changed (size, mtime or type).

        Returns a 3-tuple of sets of FileEntry objects:
        [0] added files
        [1] removed files
        [2] modified files

        self.entries is not modified; this method only reports changes.
        Candidate for making parallel
        """
        added = set()
        modified = set()
        existing_files = set()
        for dirpath_str, _, filenames in walk(str(self.path)):
            dirpath = Path(dirpath_str)
            for filename in filenames:
                # Make the assumption the database is never in the path
                abs_filename = (dirpath / filename).absolute()
                if abs_filename in self.entries:
                    entry = self.entries[abs_filename]
                    existing_files.add(entry)
                    st = lstat(str(abs_filename))
                    if entry != st:
                        modified.add(entry)
                else:
                    entry = FileEntry(self, abs_filename)
                    entry.update_attrs()
                    added.add(entry)
        removed = set(self.entries.values()) - existing_files
        return added, removed, modified

    def update(self, this_path=None):
        """
        Walks the filesystem, adding and removing files from
        the database as appropriate.

        Returns a 3-tuple of sets of filenames:
        [0] added files
        [1] removed files
        [2] modified files
        """
        if this_path is None:
            this_path = self.path
        added, removed, modified = self._find_changes()
        for entry in added:
            entry.update()  # Calculate hash
            self.entries[entry.filename] = entry
        for entry in removed:
            del self.entries[entry.filename]
        # Entries will appear in 'modified' if the size, mtime or type
        # change. This will not be reliable over time (you need hashes to do that)
        # I've seen a lot of spurious mtime mismatches on vfat
        # filesystems (like on USB flash drives), so only report files
        # as modified if the hash changes.
        content_modified = set()
        for entry in modified:
            old_entry = entry
            entry.update()
            if entry != old_entry:
                content_modified.add(entry)
        return (
            {entry.filename for entry in added},
            {entry.filename for entry in removed},
            {entry.filename for entry in content_modified},
        )

    def status(self):
        added, removed, modified = self._find_changes()
        return (
            {entry.filename for entry in added},
            {entry.filename for entry in removed},
            {entry.filename for entry in modified},
        )

    def verify(self, verbose_failures=False):
        """
        Calls each FileEntry's verify method to make sure that
        nothing has changed on disk.

        Returns a 2-tuple of sets of filenames:
        [0] modified files
        [1] removed files
        """
        modified = set()
        removed = set()
        count = len(self.entries)
        # TODO: Track number of bytes hashed instead of number of files
        # This will act as a more meaningful progress indicator
        i = 0
        for i, entry in enumerate(self.entries.values(), 1):
            if entry.exists():
                if entry.verify():
                    entry.update_attrs()
                else:
                    if verbose_failures:
                        stderr.write(
                            "\r{} failed hash verification\n".format(entry.filename)
                        )
                    modified.add(entry.filename)
            else:
                removed.add(entry.filename)
                if verbose_failures:
                    stderr.write("\r{} is missing\n".format(entry.filename))
            stderr.write("\rChecked {} of {} files".format(i, count))
        if i:
            stderr.write("\n")
        return modified, removed

    def file_paths(self):
        for entry in self.entries.values():
            yield str(entry.relative_path)

    def get_info(self):
        """Returns summary information on a file database. Uses introspection to collect statistics"""
        count_files = 0
        size_files = 0
        for entry in self.files():
            count_files += 1
            size_files += entry.size
        # Collect some stats on directory paths. This is useful to seeing if the data paths will be
        # truncated on a Joliet file path
        max_length = 0
        longest_dir = ""
        dirs = set()
        for this_dir in self.entries:
            dirs = dirs | {this_dir.parent}
            length = len(Path(this_dir).parts) - 2
            if length > max_length:
                longest_dir = this_dir.parent
                max_length = length
        # Format answer
        result = super().get_info()
        result += f"Data size       = {size_files:,} bytes\n"
        result += f"Number of dirs  = {len(dirs)}\n"
        result += f"Max dir depth   = {max_length} (on source file system)\n"
        result += f" Dir =: {longest_dir}\n"
        return result

    # Single Threaded version
    def calculate_file_hash(self, verbose=False):
        count = 0
        for entry in self.entries.values():
            entry.calculate_file_hash()
            count += 1
            if verbose:
                if (count % 1000) == 0:
                    print(f" {count}", flush=True)
                elif (count % 10) == 0:
                    print('.', end='', flush=True)
        if verbose and not ((count % 1000) == 0):  # Close off line if part finished
            print(f" {count}", flush=True)




#!/usr/bin/python3

# Copyright (C) 2017-2018  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
A drop-in replacement program for the 'genisoimage' executable that uses
PyCdlib under the hood.
"""

from __future__ import print_function

import argparse
import collections
import fileinput
import fnmatch
import os
import re
import sys
import time

try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO

import pycdlib

################################ MURMER3 HASH FUNCTIONS ##############################

if sys.version_info > (3, 0):

    def xrange(a, b, c):
        """
        An xrange that works for Python 3.
        """
        return range(a, b, c)

    def xencode(x):
        """
        A version of encode that works for bytes, bytearrays, or strings.
        """
        if isinstance(x, (bytearray, bytes)):
            return x
        return x.encode()


else:

    def xencode(x):
        """
        The identity version of xencode for Python 2.
        """
        return x


def mm3hash(key, seed=0x0):
    """ Implements 32bit murmur3 hash. """

    key = bytearray(xencode(key))

    def fmix(h):
        """
        A function to mix h.
        """
        h ^= h >> 16
        h = (h * 0x85ebca6b) & 0xFFFFFFFF
        h ^= h >> 13
        h = (h * 0xc2b2ae35) & 0xFFFFFFFF
        h ^= h >> 16
        return h

    length = len(key)
    nblocks = int(length / 4)

    h1 = seed

    c1 = 0xcc9e2d51
    c2 = 0x1b873593

    # body
    for block_start in xrange(0, nblocks * 4, 4):
        # ??? big endian?
        k1 = (
            key[block_start + 3] << 24
            | key[block_start + 2] << 16
            | key[block_start + 1] << 8
            | key[block_start + 0]
        )

        k1 = (c1 * k1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (c2 * k1) & 0xFFFFFFFF

        h1 ^= k1
        h1 = (h1 << 13 | h1 >> 19) & 0xFFFFFFFF  # inlined ROTL32
        h1 = (h1 * 5 + 0xe6546b64) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 4
    k1 = 0
    tail_size = length & 3

    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size > 0:
        k1 = (k1 * c1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (k1 * c2) & 0xFFFFFFFF
        h1 ^= k1

    # finalization
    unsigned_val = fmix(h1 ^ length)
    if unsigned_val & 0x80000000 == 0:
        return unsigned_val

    return -((unsigned_val ^ 0xFFFFFFFF) + 1)


def mm3hashfromfile(filename):
    """
    A function to generate a Murmur3 hash given a filename.
    """
    with open(filename, "rb") as infp:
        done = False
        seed = 0
        while not done:
            data = infp.read(32 * 1024)
            if len(data) < 32 * 1024:
                # EOF
                done = True
            seed = mm3hash(data, seed)

    return seed


################################ HELPER FUNCTIONS ##############################


def truncate_basename(basename, iso_level, is_dir):
    """
    A function to truncate a basename and make it conformant to the passed-in
    ISO interchange level.

    Parameters:
     basename - The initial basename to truncate and translate
     iso_level - The ISO interchange level to follow when truncating/translating
     is_dir - Whether this is a directory or a file
    Returns:
     The truncated and translated name suitable for the ISO interchange level
     specified.
    """
    if iso_level == 4:
        # ISO level 4 allows "anything", so just return the original.
        return basename

    if iso_level == 1:
        maxlen = 8
    else:
        maxlen = 31 if is_dir else 30

    # For performance reasons, we first truncate the string to the length
    # allowed.  Second, ISO9660 Levels 1, 2, and 3 require all uppercase names,
    # so we uppercase it.
    valid_base = basename[:maxlen].upper()

    # Finally, ISO9660 requires only uppercase letters, 0-9, and underscore.
    # Translate any non-compliant characters to underscore and return that.
    return re.sub("[^A-Z0-9_]{1}", r"_", valid_base)


def mangle_file_for_iso9660(orig, iso_level):
    """
    A function to take a regular Unix-style filename (including extension) and
    produce a tuple consisting of an ISO9660-valid basename and an ISO9660-valid
    extension.

    Parameters:
     orig - The original filename
     iso_level - The ISO interchange level to conform to
    Returns:
     A tuple where the first entry is the ISO9660-compliant basename and where
     the second entry is the ISO9660-compliant extension.
    """
    # ISO9660 has a lot of restrictions on what valid names are.  Here, we mangle
    # the names to conform to those rules.  In particular, the rules for filenames are:
    # 1.  Filenames can only consist of d-characters or d1-characters; these are defined
    #     in the Appendix as: 0-9A-Z_
    # 2.  Filenames look like:
    #     - zero or more d-characters (filename)
    #     - separator 1 (.)
    #     - zero or more d-characters (extension)
    #     - separate 2 (;)
    #     - version, between 0 and 32767
    # If the filename contains zero characters, then the extension must contain at least
    # one character, and vice versa.
    # 3.  If this is iso level one, then the length of the filename cannot exceed 8 and
    #     the length of the extension cannot exceed 3.  In levels 2 and 3, the length of
    #     the filename+extension cannot exceed 30.
    #
    # This function takes any valid Unix filename and converts it into one that is allowed
    # by the above rules.  It does this by substituting _ for any invalid characters in
    # the filename, and by shortening the name to a form of aaa_xxxx.eee;1 (if necessary).
    # The aaa is always the first three characters of the original filename; the xxxx is
    # the next number in a sequence starting from 0.

    valid_ext = ""
    splitter = orig.split(".")
    if iso_level == 4:
        # A level 4 ISO allows 'anything', so just return the original.
        if len(splitter) == 1:
            return orig, valid_ext

        ext = splitter[-1]
        return orig[: len(orig) - len(ext) - 1], ext

    if len(splitter) == 1:
        # No extension specified, leave ext empty
        basename = orig
    else:
        ext = splitter[-1]
        basename = orig[: len(orig) - len(ext) - 1]

        # If the extension is empty, too long (> 3), or contains any illegal characters,
        # we treat it as part of the basename instead
        extlen = len(ext)
        if extlen == 0 or extlen > 3:
            valid_ext = ""
            basename = orig
        else:
            tmpext = ext.upper()
            valid_ext, numsub = re.subn("[^A-Z0-9_]{1}", r"_", tmpext)
            if numsub > 0:
                valid_ext = ""
                basename = orig

    # All right, now we have the basename of the file, and (optionally) an extension.
    return truncate_basename(basename, iso_level, False), valid_ext + ";1"


def mangle_dir_for_iso9660(orig, iso_level):
    """
    A function to take a regular Unix-style directory name and produce an
    ISO9660-valid directory name.

    Parameters:
     orig - The original filename
     iso_level - The ISO interchange level to conform to
    Returns:
     An ISO9660-compliant directory name.
    """
    # ISO9660 has a lot of restrictions on what valid directory names are.  Here, we mangle
    # the names to conform to those rules.  In particular, the rules for dirnames are:
    # 1.  Filenames can only consist of d-characters or d1-characters; these are defined
    #     in the Appendix as: 0-9A-Z_
    # 2.  If this is ISO level one, then directory names consist of no more than 8 characters
    # This function takes any valid Unix directory name and converts it into one that is
    # allowed by the above rules.  It does this by substituting _ for any invalid character
    # in the directory name, and by shortening the name to a form of aaaaxxx (if necessary).
    # The aaa is always the first three characters of the original filename; the xxxx is
    # the next number in a sequence starting from 0.

    return truncate_basename(orig, iso_level, True)


def match_entry_to_list(pattern_list, entry):
    """
    A function to match a string to a list of filename patterns.  If any of them
    match, returns True, otherwise, returns False.

    Parameters:
     pattern_list - The list of filename patterns to check against
     entry - The string to check
    Returns:
     True if the string matches any of the filename patterns, False otherwise.
    """
    for pattern in pattern_list:
        if fnmatch.fnmatch(entry, pattern):
            return True

    return False


def parse_file_list(thelist):
    """
    A function to take a list of filenames, open each one for reading, and
    yield each of the lines in the file.

    Parameters:
     thelist - The list of files to open
    Returns:
     Nothing.
    """
    for f in thelist:
        with open(f, "r") as infp:
            for line in infp.xreadlines():
                yield line.rstrip()


def build_joliet_path(root, name):
    """
    A function to build a complete, valid Joliet path based on a root directory
    and a name.

    Parameters:
     root - The root directory name
     name - The name for the Joliet entry
    Returns:
     A valid, absolute Joliet filename.
    """
    if root and root[0] == "/":
        root = root[1:]
    intermediate = ""
    for intdir in root.split("/"):
        if not intdir:
            continue

        intermediate += "/" + intdir[:64]

    return intermediate + "/" + name[:64]


def build_udf_path(root, name):
    """
    A function to build a complete, valid UDF path based on a root directory
    and a name.

    Parameters:
     root - The root directory name.
     name - The name for the UDF entry.
    Returns:
     A valid, absolute UDF filename.
    """
    if root and root[0] == "/":
        root = root[1:]
    intermediate = ""
    for intdir in root.split("/"):
        if not intdir:
            continue

        intermediate += "/" + intdir

    return intermediate + "/" + name


class EltoritoEntry(object):
    """
    A class that represents a single El Torito entry on the ISO.  There may be
    more than one of these per-ISO, so each one is tracked separately.
    """

    __slots__ = (
        "bootfile_orig",
        "bootfile_parts",
        "mediatype",
        "boot",
        "load_size",
        "load_seg",
        "boot_info_table",
        "bootfile_iso_path",
        "dirlevel",
    )

    def __init__(self):
        self.bootfile_orig = ""
        self.bootfile_parts = []
        self.mediatype = "floppy"
        self.boot = True
        self.load_size = 0
        self.load_seg = 0
        self.boot_info_table = False
        self.bootfile_iso_path = ""
        self.dirlevel = None

    def set_bootfile(self, bootfile):
        """
        A method to set the bootfile for this El Torito entry.

        Parameters:
         bootfile - The bootfile to set for this El Torito entry
        Returns:
         Nothing.
        """
        self.bootfile_orig = bootfile
        self.bootfile_parts = bootfile.split("/")


def parse_arguments():
    """
    A function to parse all of the arguments passed to the executable.  Note
    that the set of arguments accepted is intentionally compatible with
    genisoimage.  Also note that this executable does not support all flags
    below, and will silently ignore ones it doesn't understand.  This is to
    keep maximum compatibility with genisoimage.

    Parameters:
     None.
    Returns:
     An ArgumentParser object with the parsed command-line arguments.
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-nobak", "-no-bak", help="Do not include backup files", action="store_true"
    )
    parser.add_argument(
        "-abstract", help="Set Abstract filename", action="store", default=""
    )
    parser.add_argument(
        "-appid", "-A", help="Set Application ID", action="store", default=""
    )
    parser.add_argument(
        "-biblio", help="Set Bibliographic filename", action="store", default=""
    )
    parser.add_argument(
        "-cache-inodes",
        help="Cache inodes (needed to detect hard links)",
        action="store_true",
    )
    parser.add_argument(
        "-no-cache-inodes",
        help="Do not cache inodes (if filesystem has no unique unides)",
        action="store_true",
    )
    parser.add_argument(
        "-check-oldnames",
        help="Check all imported ISO9660 names from old session",
        action="store_true",
    )
    parser.add_argument(
        "-check-session",
        help="Check all ISO9660 names from previous session",
        action="store",
    )
    parser.add_argument(
        "-copyright", help="Set Copyright filename", action="store", default=""
    )
    parser.add_argument("-debug", help="Set debug flag", action="store_true")
    parser.add_argument(
        "-eltorito-boot", "-b", help="Set El Torito boot image name", action="store"
    )
    parser.add_argument(
        "-efi-boot", "-e", help="Set EFI boot image name", action="append"
    )
    parser.add_argument(
        "-eltorito-alt-boot",
        help="Start specifying alternative El Torito boot parameters",
        action="append_const",
        const=True,
    )
    parser.add_argument(
        "-sparc-boot", "-B", help="Set sparc boot image names", action="store"
    )
    parser.add_argument(
        "-sunx86-boot", help="Set sunx86 boot image names", action="store"
    )
    parser.add_argument(
        "-generic-boot", "-G", help="Set generic boot image name", action="store"
    )
    parser.add_argument(
        "-sparc-label", help="Set sparc boot disk label", action="store", nargs=2
    )
    parser.add_argument(
        "-sunx86-label", help="Set sunx86 boot disk label", action="store", nargs=2
    )
    parser.add_argument(
        "-eltorito-catalog",
        "-c",
        help="Set El Torito boot catalog name",
        action="store",
        default=None,
    )
    parser.add_argument(
        "-cdrecord-params", "-C", help="Magic parameters from cdrecord", action="store"
    )
    parser.add_argument(
        "-omit-period",
        "-d",
        help="Omit trailing periods from filenames (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-dir-mode", help="Make the mode of all directories this mode", action="store"
    )
    parser.add_argument(
        "-disable-deep-relocation",
        "-D",
        help="Disable deep directory relocation (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-file-mode", help="Make the mode of all plain files this mode", action="store"
    )
    parser.add_argument(
        "-follow-links", "-f", help="Follow symbolic links", action="store_true"
    )
    parser.add_argument(
        "-gid", help="Make the group owner of all files this gid", action="store"
    )
    parser.add_argument(
        "-graft-points",
        help="Allow to use graft points for filenames",
        action="store_true",
    )
    parser.add_argument(
        "-root",
        help="Set root directory for all new files and directories",
        action="store",
    )
    parser.add_argument(
        "-old-root",
        help="Set root directory in previous session this is searched for files",
        action="store",
    )
    parser.add_argument("-help", help="Print option help", action="help")
    parser.add_argument(
        "-hide", help="Hide ISO9660/RR file", action="append", default=[]
    )
    parser.add_argument(
        "-hide-list",
        help="File with list of ISO9660/RR files to hide",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-hidden",
        help="Set hidden attribute on ISO9660 file",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-hidden-list",
        help="File with list of ISO9660 files with hidden attribute",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-hide-joliet", help="Hide Joliet file", action="append", default=[]
    )
    parser.add_argument(
        "-hide-joliet-list",
        help="File with list of Joliet files to hide",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-hide-joliet-trans-tbl",
        help="Hide TRANS.TBL from Joliet tree",
        action="store_true",
    )
    parser.add_argument(
        "-hide-rr-moved",
        help="Rename RR_MOVED to .rr_moved in Rock Ridge tree",
        action="store_true",
    )
    parser.add_argument("-gui", help="Switch behavior for GUI", action="store_true")
    parser.add_argument("-i", help="No longer supported", action="store")
    parser.add_argument(
        "-input-charset",
        help="Local input charset for file name conversion",
        action="store",
    )
    parser.add_argument(
        "-output-charset",
        help="Output charset for file name conversion",
        action="store",
    )
    parser.add_argument(
        "-iso-level",
        help="Set ISO9660 conformance level (1..3) or 4 for ISO9660 version 2",
        action="store",
        default=1,
        type=int,
        choices=range(1, 5),
    )
    parser.add_argument(
        "-joliet",
        "-J",
        help="Generate Joliet directory information",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-joliet-long",
        help="Allow Joliet file names to be 103 Unicode characters",
        action="store_true",
    )
    parser.add_argument(
        "-jcharset",
        help="Local charset for Joliet directory information",
        action="store",
    )
    parser.add_argument(
        "-full-iso9660-filenames",
        "-l",
        help="Allow full 31 character filenames for ISO9660 names",
        action="store_true",
    )
    parser.add_argument(
        "-max-iso9660-filenames",
        help="Allow 37 character filenames for ISO9660 names (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-allow-limited-size",
        help="Allow different file sizes in ISO9660/UDF on large files",
        action="store_true",
    )
    parser.add_argument(
        "-allow-leading-dots",
        "-ldots",
        "-L",
        help="Allow ISO9660 filenames to start with '.' (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-log-file", help="Re-direct messages to LOG_FILE", action="store"
    )
    parser.add_argument(
        "-exclude", "-m", help="Exclude file name", action="append", default=[]
    )
    parser.add_argument(
        "-exclude-list",
        help="File with list of file names to exclude",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-pad", help="Pad output to a multiple of 32k (default)", action="store_true"
    )
    parser.add_argument(
        "-no-pad", help="Do not pad output to a multiple of 32k", action="store_true"
    )
    parser.add_argument(
        "-prev-session",
        "-M",
        help="Set path to previous session to merge",
        action="store",
    )
    parser.add_argument("-dev", help="Device", action="store")
    parser.add_argument(
        "-omit-version-number",
        "-N",
        help="Omit version number from ISO9660 filename (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-new-dir-mode", help="Mode used when creating new directories", action="store"
    )
    parser.add_argument(
        "-force-rr",
        help="Inhibit automatic Rock Ridge detection for previous session",
        action="store_true",
    )
    parser.add_argument(
        "-no-rr",
        help="Inhibit reading of Rock Ridge attributes from previous session",
        action="store_true",
    )
    parser.add_argument(
        "-no-split-symlink-components",
        help="Inhibit splitting symlink components",
        action="store_true",
    )
    parser.add_argument(
        "-no-split-symlink-fields",
        help="Inhibit splitting symlink fields",
        action="store_true",
    )
    parser.add_argument("-output", "-o", help="Set output file name", action="store")
    parser.add_argument(
        "-path-list", help="File with list of pathnames to process", action="store"
    )
    parser.add_argument(
        "-preparer", "-p", help="Set Volume preparer", action="store", default=""
    )
    parser.add_argument(
        "-print-size",
        help="Print estimated filesystem size and exit",
        action="store_true",
    )
    parser.add_argument(
        "-publisher", "-P", help="Set Volume publisher", action="store", default=""
    )
    parser.add_argument("-quiet", help="Run quietly", action="store_true")
    parser.add_argument(
        "-rational-rock",
        "-r",
        help="Generate rationalized Rock Ridge directory information",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-rock",
        "-R",
        help="Generate Rock Ridge directory information",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-sectype",
        "-s",
        help="Set output sector type to e.g. data/xa1/raw",
        action="store",
    )
    parser.add_argument(
        "-alpha-boot",
        help="Set alpha boot image name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-hppa-cmdline",
        help="Set hppa boot command line (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-hppa-kernel-32",
        help="Set hppa 32-bit image name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-hppa-kernel-64",
        help="Set hppa 64-bit image name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-hppa-bootloader",
        help="Set hppa boot loader file name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-hppa-ramdisk",
        help="Set hppa ramdisk file name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-mips-boot",
        help="Set mips boot image name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-mipsel-boot",
        help="Set mipsel boot image name (relative to image root)",
        action="store",
    )
    parser.add_argument(
        "-jigdo-jigdo",
        help="Produce a jigdo .jigdo file as well as the .iso",
        action="store",
    )
    parser.add_argument(
        "-jigdo-template",
        help="Produce a jigdo .template file as well as the .iso",
        action="store",
    )
    parser.add_argument(
        "-jigdo-min-file-size",
        help="Minimum size for a file to be listed in the jigdo file",
        action="store",
    )
    parser.add_argument(
        "-jigdo-force-md5",
        help="Pattern(s) where files MUST match an externally-supplied MD5Sum",
        action="store",
    )
    parser.add_argument(
        "-jigdo-exclude",
        help="Pattern(s) to exclude from the jigdo file",
        action="store",
    )
    parser.add_argument(
        "-jigdo-map",
        help="Pattern(s) to map paths (e.g. Debian=/mirror/debian)",
        action="store",
    )
    parser.add_argument(
        "-md5-list",
        help="File containing MD5 sums of the files that should be checked",
        action="store",
    )
    parser.add_argument(
        "-jigdo-template-compress",
        help="Choose to use gzip or bzip2 compression for template data; default is gzip",
        action="store",
    )
    parser.add_argument(
        "-checksum_algorithm_iso",
        help="Specify the checksum types desired for the output image",
        action="store",
    )
    parser.add_argument(
        "-checksum_algorithm_template",
        help="Specify the checksum types desired for the output jigdo template",
        action="store",
    )
    parser.add_argument(
        "-sort",
        help="Sort file content locations according to rules in FILE",
        action="store",
    )
    parser.add_argument(
        "-split-output",
        help="Split output into files of approx. 1GB size",
        action="store_true",
    )
    parser.add_argument(
        "-stream-file-name",
        help="Set the stream file ISO9660 name (incl. version)",
        action="store",
    )
    parser.add_argument(
        "-stream-media-size",
        help="Set the size of your CD media in sectors",
        action="store",
    )
    parser.add_argument("-sysid", help="Set System ID", action="store", default="")
    parser.add_argument(
        "-translation-table",
        "-T",
        help="Generate translation tables for systems that don't understand long filenames",
        action="store_true",
    )
    parser.add_argument(
        "-table-name", help="Translation table file name", action="store"
    )
    parser.add_argument(
        "-ucs-level", help="Set Joliet UCS level (1..3)", action="store"
    )
    parser.add_argument("-udf", help="Generate UDF file system", action="store_true")
    parser.add_argument(
        "-dvd-video",
        help="Generate DVD-Video compliant UDF file system",
        action="store_true",
    )
    parser.add_argument(
        "-uid", help="Make the owner of all files this uid", action="store"
    )
    parser.add_argument(
        "-untranslated-filenames",
        "-U",
        help="Allow Untranslated filenames (for HPUX & AIX - violates ISO9660).  Forces -l, -d, -N, -allow-leading-dots, -relaxed-filenames, -allow-lowercase, -allow-multidot",
        action="store_true",
    )
    parser.add_argument(
        "-relaxed-filenames",
        help="Allow 7 bit ASCII except lower case characters (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-no-iso-translate",
        help="Do not translate illegal ISO characters '~', '-', and '#' (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-allow-lowercase",
        help="Allow lower case characters in addition to the current character set (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-allow-multidot",
        help="Allow more than one dot in filenames (e.g. .tar.gz) (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-use-fileversion", help="Use fileversion # from filesystem", action="store"
    )
    parser.add_argument("-verbose", "-v", help="Verbose", action="store_true")
    parser.add_argument(
        "-version", help="Print the current version", action="store_true"
    )
    parser.add_argument(
        "-volid", "-V", help="Set Volume ID", action="store", default=""
    )
    parser.add_argument("-volset", help="Set Volume set ID", action="store", default="")
    parser.add_argument(
        "-volset-size", help="Set Volume set size", action="store", default=1
    )
    parser.add_argument(
        "-volset-seqno",
        help="Set Volume set sequence number",
        action="store",
        default=1,
    )
    parser.add_argument(
        "-old-exclude",
        "-x",
        help="Exclude file name (deprecated)",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-hard-disk-boot",
        help="Boot image is a hard disk image",
        action="append_const",
        const=True,
    )
    parser.add_argument(
        "-no-emul-boot",
        help="Boot image is a 'no emulation' image",
        action="append_const",
        const=True,
    )
    parser.add_argument(
        "-no-boot", help="Boot image is not bootable", action="append_const", const=True
    )
    parser.add_argument(
        "-boot-load-seg", help="Set load segment for boot image", action="append"
    )
    parser.add_argument(
        "-boot-load-size", help="Set number of load sectors", action="append"
    )
    parser.add_argument(
        "-boot-info-table",
        help="Patch boot image with info table",
        action="append_const",
        const=True,
    )
    parser.add_argument(
        "-XA", help="Generate XA directory attributes", action="store_true"
    )
    parser.add_argument(
        "-xa", help="Generate rationalized XA directory attributes", action="store_true"
    )
    parser.add_argument(
        "-transparent-compression",
        "-z",
        help="Enable transparent compression of files",
        action="store_true",
    )
    parser.add_argument("-hfs-type", help="Set HFS default TYPE", action="store")
    parser.add_argument("-hfs-creator", help="Set HFS default CREATOR", action="store")
    parser.add_argument(
        "-apple", "-g", help="Add Apple ISO9660 extensions", action="store_true"
    )
    parser.add_argument(
        "-hfs", "-h", help="Create ISO9660/HFS hybrid", action="store_true"
    )
    parser.add_argument(
        "-map", "-H", help="Map file extensions to HFS TYPE/CREATOR", action="store"
    )
    parser.add_argument(
        "-magic", help="Magic file for HFS TYPE/CREATOR", action="store"
    )
    parser.add_argument(
        "-probe", help="Probe all files for Apple/Unix file types", action="store_true"
    )
    parser.add_argument(
        "-mac-name",
        help="Use Macintosh name for ISO9660/Joliet/RockRidge file name",
        action="store_true",
    )
    parser.add_argument(
        "-no-mac-files",
        help="Do not look for Unix/Mac files (deprecated)",
        action="store_true",
    )
    parser.add_argument(
        "-boot-hfs-file", help="Set HFS boot image name", action="store"
    )
    parser.add_argument(
        "-part", help="Generate HFS partition table", action="store_true"
    )
    parser.add_argument(
        "-cluster-size",
        help="Cluster size for PC Exchange Macintosh files",
        action="store",
    )
    parser.add_argument("-auto", help="Set HFS AutoStart file name", action="store")
    parser.add_argument(
        "-no-desktop",
        help="Do not create the HFS (empty) Desktop files",
        action="store_true",
    )
    parser.add_argument("-hide-hfs", help="Hide HFS file", action="append", default=[])
    parser.add_argument(
        "-hide-hfs-list", help="List of HFS files to hide", action="append", default=[]
    )
    parser.add_argument(
        "-hfs-volid", help="Volume name for the HFS partition", action="store"
    )
    parser.add_argument(
        "-icon-position", help="Keep HFS icon position", action="store_true"
    )
    parser.add_argument("-root-info", help="finderinfo for root folder", action="store")
    parser.add_argument(
        "-input-hfs-charset",
        help="Local input charset for HFS file name conversion",
        action="store",
    )
    parser.add_argument(
        "-output-hfs-charset",
        help="Output charset for HFS file name conversion",
        action="store",
    )
    parser.add_argument(
        "-hfs-unlock", help="Leave HFS volume unlocked", action="store_true"
    )
    parser.add_argument(
        "-hfs-bless", help="Name of Folder to be blessed", action="store"
    )
    parser.add_argument(
        "-hfs-parms", help="Comma separated list of HFS parameters", action="store"
    )
    parser.add_argument(
        "-prep-boot", help="PReP boot image file -- up to 4 are allowed", action="store"
    )  # FIXME: we need to allow between 1 and 4 arguments
    parser.add_argument("-chrp-boot", help="Add CHRP boot header", action="store_true")
    parser.add_argument(
        "--cap", help="Look for AUFS CAP Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--netatalk", help="Look for NETATALK Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--double", help="Look for AppleDouble Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--ethershare",
        help="Look for Helios EtherShare Macintosh files",
        action="store_true",
    )
    parser.add_argument(
        "--exchange", help="Look for PC Exchange Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--sgi", help="Look for SGI Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--macbin", help="Look for MacBinary Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--single", help="Look for AppleSingle Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--ushare", help="Look for IPT UShare Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--xinet", help="Look for XINET Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--dave", help="Look for DAVE Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--sfm", help="Look for SFM Macintosh files", action="store_true"
    )
    parser.add_argument(
        "--osx-double",
        help="Look for MacOS X AppleDouble Macintosh files",
        action="store_true",
    )
    parser.add_argument(
        "--osx-hfs", help="Look for MacOS X HFS Macintosh files", action="store_true"
    )
    parser.add_argument(
        "-find",
        help="Option separator: Use find command line to the right",
        action="store",
    )
    parser.add_argument(
        "-posix-H",
        help="Follow symbolic links encountered on command line",
        action="store_true",
    )
    parser.add_argument(
        "-posix-L", help="Follow all symbolic links", action="store_true"
    )
    parser.add_argument(
        "-posix-P", help="Do not follow symbolic links (default)", action="store_true"
    )
    parser.add_argument(
        "-rrip110", help="Create old Rock Ridge V 1.10", action="store_true"
    )
    parser.add_argument(
        "-rrip112", help="Create new Rock Ridge V 1.12 (default)", action="store_true"
    )
    parser.add_argument("-ignore-error", help="Ignore errors", action="store_true")
    parser.add_argument(
        "-eltorito-platform",
        help="Set El Torito platform id for the next boot entry",
        action="store",
    )
    parser.add_argument(
        "-data-change-warn",
        help="Treat data/size changes as warning only",
        action="store_true",
    )
    parser.add_argument(
        "-errctl", help="Read error control defs from file or inline.", action="store"
    )
    parser.add_argument("-hide-udf", help="Hide UDF file", action="append", default=[])
    parser.add_argument(
        "-hide-udf-list",
        help="File with list of UDF files to hide",
        action="append",
        default=[],
    )
    parser.add_argument(
        "-long-rr-time", help="Use long Rock Ridge time format", action="store_true"
    )
    parser.add_argument(
        "-modification-date",
        help="Set the modification date field of the PVD",
        action="store",
    )
    parser.add_argument(
        "-no-limit-pathtables",
        help="Allow more than 65535 parent directories (violates ISO9660)",
        action="store_true",
    )
    parser.add_argument(
        "-no-long-rr-time",
        "-short-rr-time",
        help="Use short Rock Ridge time format",
        action="store_true",
    )
    parser.add_argument("-UDF", help="Generate UDF file system", action="store_true")
    parser.add_argument(
        "-udf-symlinks",
        help="Create symbolic links on UDF image (default)",
        action="store_true",
    )
    parser.add_argument(
        "-no-udf-symlinks",
        help="Do not create symbolic links on UDF image",
        action="store_true",
    )
    parser.add_argument(
        "-no-hfs", help="Do not create ISO9660/HFS hybrid", action="store_true"
    )
    parser.add_argument(
        "-scan-for-duplicates",
        help="Aggressively try to find duplicate files to reduce size (very slow!)",
        action="store_true",
    )
    parser.add_argument(
        "paths", help="Paths to get data from", action="store", nargs=argparse.REMAINDER
    )
    return parser.parse_args()


def determine_eltorito_entries(args):
    """
    A function to build up the list of EltoritoEntry objects for this ISO.

    Parameters:
     args - The ArgumentParser object returned from parse_arguments()
    Returns:
     A list of EltoritoEntry objects for this ISO (which may be empty).
    """
    eltorito_entries = []
    efi_boot_index = 0
    load_seg_index = 0
    load_size_index = 0

    for arg in sys.argv[1:]:
        if arg == "-eltorito-alt-boot":
            eltorito_entries.append(EltoritoEntry())
            continue

        if arg not in (
            "-b",
            "-eltorito-boot",
            "-e",
            "-efi-boot",
            "-no-emul-boot",
            "-hard-disk-boot",
            "-no-boot",
            "-boot-load-seg",
            "-boot-load-size",
            "-boot-info-table",
        ):
            continue

        if not eltorito_entries:
            entry = EltoritoEntry()
            eltorito_entries.append(entry)
        else:
            entry = eltorito_entries[-1]

        if arg in ("-b", "-eltorito-boot"):
            entry.set_bootfile(args.eltorito_boot)
        elif arg in ("-e", "-efi-boot"):
            entry.set_bootfile(args.efi_boot[efi_boot_index])
            efi_boot_index += 1
        elif arg == "-no-emul-boot":
            entry.mediatype = "noemul"
        elif arg == "-hard-disk-boot":
            entry.mediatype = "hdemul"
        elif arg == "-no-boot":
            entry.boot = False
        elif arg == "-boot-load-seg":
            entry.load_seg = int(args.boot_load_seg[load_seg_index])
            load_seg_index += 1
        elif arg == "-boot-load-size":
            entry.load_size = int(args.boot_load_size[load_size_index])
            load_size_index += 1
        elif arg == "-boot-info-table":
            entry.boot_info_table = True

    return eltorito_entries


class DirLevel(object):
    """
    A class to hold information about one directory level of the directory
    hierarchy.  Each level has an iso_path, a joliet_path, and a set of
    mangled maps for mangling filenames as appropriate.
    """

    __slots__ = ("iso_path", "joliet_path", "udf_path", "mangled_children")

    def __init__(self, iso_path, joliet_path, udf_path):
        self.iso_path = iso_path
        self.joliet_path = joliet_path
        self.udf_path = udf_path
        self.mangled_children = {}


def build_iso_path(parent_dirlevel, nameonly, iso_level, is_dir):
    """
    A function to build an absolute ISO path from a DirLevel object, a name,
    an ISO interchange level, and whether this is a directory or not.

    Parameters:
     parent_dirlevel - The DirLevel object representing the parent.
     nameonly - The basename of the ISO path.
     iso_level - The ISO interchange level to use.
     is_dir - Whether this is a directory or not.
    Returns:
     A string representing the ISO absolute path.
    """

    # Mangling the name and keeping track is a complicated affair.  First off,
    # we mangle the incoming dirname so it conforms to ISO9660 rules (see
    # mangle_dir_for_iso9660() for more info on that).  Once we have the
    # mangled name, we see if that name has been used at this directory level
    # yet.  If it has not been used, we mark it as now used, and return it
    # unmolested (beyond the ISO9660 mangle).  If it has been used, then we
    # need to strip it down to its prefix (the first 4 characters), add a
    # 3-digit number starting at zero, then iterate until we find a free one.
    # Once we have found a free one, we mark it as now used, and return what
    # we figured out.

    if is_dir:
        filemangle = mangle_dir_for_iso9660(nameonly, iso_level)
    else:
        filename, ext = mangle_file_for_iso9660(nameonly, iso_level)
        if ext == "":
            filemangle = filename
        else:
            filemangle = ".".join([filename, ext])

    if filemangle in parent_dirlevel.mangled_children:
        currnum = 0
        prefix = filemangle[:5]
        while True:
            if is_dir:
                tmp = "%s%.03d" % (prefix, currnum)
            else:
                tmp = "%s%.03d.%s" % (prefix, currnum, ext)
            if tmp not in parent_dirlevel.mangled_children:
                filemangle = tmp
                break
            currnum += 1
            if currnum == 1000:
                return None

    parent_dirlevel.mangled_children[filemangle] = True

    parent_iso_path = parent_dirlevel.iso_path
    if parent_dirlevel.iso_path == "/":
        parent_iso_path = parent_dirlevel.iso_path[1:]
    return parent_iso_path + "/" + filemangle


################################### MAIN #######################################


def main():
    """
    The main function for this executable that does the main work of generating
    an ISO given the parameters specified by the user.
    """
    args = parse_arguments()

    eltorito_entries = determine_eltorito_entries(args)

    if args.quiet:
        logfp = open(os.devnull, "w")
    else:
        if args.log_file is not None:
            print("re-directing all messages to %s" % (args.log_file))
            logfp = open(args.log_file, "w")
        else:
            logfp = sys.stdout

    print("pycdlib-genisoimage 1.0.0", file=logfp)

    # Check out all of the arguments we can here.
    if args.version:
        sys.exit(0)

    rock_version = None
    if args.rational_rock or args.rock:
        rock_version = "1.09"
    if args.rrip112:
        rock_version = "1.12"

    udf_version = None
    if args.udf or args.UDF:
        udf_version = "2.60"

    if args.joliet and rock_version is None:
        print(
            "Warning: creating filesystem with Joliet extensions but without Rock Ridge",
            file=logfp,
        )
        print(
            "         extensions. It is highly recommended to add Rock Ridge.",
            file=logfp,
        )

    if args.eltorito_catalog is not None and not eltorito_entries:
        print("genisoimage: No boot image specified.", file=logfp)
        sys.exit(255)

    if args.i is not None:
        print("genisoimage: -i option no longer supported.", file=logfp)
        sys.exit(255)

    hidden_patterns = args.hidden
    for pattern in parse_file_list(args.hidden_list):
        hidden_patterns.append(pattern)

    exclude_patterns = args.exclude + args.old_exclude
    for pattern in parse_file_list(args.exclude_list):
        exclude_patterns.append(pattern)

    hide_patterns = args.hide
    for pattern in parse_file_list(args.hide_list):
        hide_patterns.append(pattern)

    hide_joliet_patterns = args.hide_joliet
    for pattern in parse_file_list(args.hide_joliet_list):
        hide_joliet_patterns.append(pattern)

    hide_udf_patterns = args.hide_udf
    for pattern in parse_file_list(args.hide_udf_list):
        hide_udf_patterns.append(pattern)

    ignore_patterns = []
    if args.nobak:
        ignore_patterns.extend(("*~*", "*#*", "*.bak"))

    if args.print_size:
        fp = BytesIO()
    else:
        if args.output is None:
            print("Output file must be specified (use -o)", file=logfp)
            sys.exit(1)

        fp = open(args.output, "wb")

    # Figure out Joliet flag, which is the combination of args.joliet
    # and args.ucs_level.
    joliet_level = None
    if args.joliet:
        joliet_level = 3
        if args.ucs_level is not None:
            joliet_level = int(args.ucs_level)

    eltorito_catalog_path = ""
    eltorito_catalog_parts = []
    if args.eltorito_catalog is not None:
        eltorito_catalog_parts = args.eltorito_catalog.split("/")

    # Create a new PyCdlib object.
    iso = pycdlib.PyCdlib()

    if args.hide_rr_moved:
        iso.set_relocated_name("_RR_MOVE", ".rr_moved")

    # Create a new ISO.
    iso.new(
        interchange_level=args.iso_level,
        sys_ident=args.sysid,
        vol_ident=args.volid,
        set_size=args.volset_size,
        seqnum=args.volset_seqno,
        vol_set_ident=args.volset,
        pub_ident_str=args.publisher,
        preparer_ident_str=args.preparer,
        app_ident_str=args.appid,
        copyright_file=args.copyright,
        abstract_file=args.abstract,
        bibli_file=args.biblio,
        joliet=joliet_level,
        rock_ridge=rock_version,
        xa=(args.XA or args.xa),
        udf=udf_version,
    )

    path_list = args.paths

    if args.path_list is not None:
        for line in fileinput.input(args.path_list):
            path_list.append(line.strip())

    seen_hashes = {}
    for path in path_list:
        check_eltorito_catalog = len(eltorito_catalog_parts) > 0
        root_level = DirLevel("/", "/", "/")
        for eltorito_entry in eltorito_entries:
            eltorito_entry.dirlevel = root_level
        entries = collections.deque(
            [(os.path.normpath(path), root_level, False, check_eltorito_catalog)]
        )
        while entries:
            localpath, parent_level, add_dir, check_eltorito_catalog = entries.popleft()
            basename = os.path.basename(localpath)

            if check_eltorito_catalog and len(eltorito_catalog_parts) == 1:
                filename, ext = mangle_file_for_iso9660(
                    eltorito_catalog_parts.pop(), args.iso_level
                )
                eltorito_catalog_path += "/" + filename + "." + ext

            for eltorito_entry in eltorito_entries:
                if (
                    eltorito_entry.dirlevel == parent_level
                    and len(eltorito_entry.bootfile_parts) == 1
                ):
                    filename, ext = mangle_file_for_iso9660(
                        eltorito_entry.bootfile_parts.pop(), args.iso_level
                    )
                    tail = "." + ext
                    if ext == "":
                        tail = ""
                    eltorito_entry.bootfile_iso_path += "/" + filename + tail

            rr_name = None
            if args.rational_rock or args.rock:
                rr_name = basename

            joliet_path = None
            if args.joliet:
                joliet_path = build_joliet_path(parent_level.joliet_path, basename)

            udf_path = None
            if args.udf or args.UDF:
                udf_path = build_udf_path(parent_level.udf_path, basename)

            if os.path.islink(localpath):
                if (not args.rational_rock or args.rock) and (not args.udf or args.UDF):
                    print("Symlink %s ignored - continuing." % (localpath), file=logfp)
                else:
                    iso_path = build_iso_path(
                        parent_level, basename, args.iso_level, False
                    )
                    if iso_path is None:
                        print(
                            "Could not find free ISO9660 name for path %s; skipping"
                            % (localpath),
                            file=logfp,
                        )

                    rr_target = None
                    if args.rational_rock or args.rock:
                        rr_target = os.readlink(localpath)

                    udf_target = None
                    if args.udf or args.UDF:
                        udf_target = os.readlink(localpath)

                    iso.add_symlink(
                        iso_path,
                        rr_symlink_name=rr_name,
                        rr_path=rr_target,
                        udf_symlink_path=udf_path,
                        udf_target=udf_target,
                        joliet_path=joliet_path,
                    )

            elif os.path.isdir(localpath):
                if add_dir:
                    iso_path = build_iso_path(
                        parent_level, basename, args.iso_level, True
                    )
                    if iso_path is None:
                        print(
                            "Could not find free ISO9660 name for path %s; skipping"
                            % (localpath),
                            file=logfp,
                        )
                    depth = iso_path.count("/")
                    if rr_name is None and depth > 7:
                        print(
                            "Directories too deep for '%s' (%d) max is 7; ignored - continuing."
                            % (localpath, depth),
                            file=logfp,
                        )
                        continue
                    iso.add_directory(
                        iso_path,
                        rr_name=rr_name,
                        joliet_path=joliet_path,
                        udf_path=udf_path,
                    )
                else:
                    iso_path = parent_level.iso_path
                    joliet_path = parent_level.joliet_path
                    udf_path = parent_level.udf_path

                on_eltorito_catalog_path = False
                eltorito_duplicate_check = ""
                if (
                    check_eltorito_catalog
                    and len(eltorito_catalog_parts) > 1
                    and eltorito_catalog_parts[0] == basename
                ):
                    eltorito_catalog_path += "/" + mangle_dir_for_iso9660(
                        basename, args.iso_level
                    )
                    eltorito_catalog_parts.pop(0)
                    on_eltorito_catalog_path = True
                    if len(eltorito_catalog_parts) == 1:
                        eltorito_duplicate_check = eltorito_catalog_parts[0]

                parent = DirLevel(iso_path, joliet_path, udf_path)
                for eltorito_entry in eltorito_entries:
                    if (
                        parent_level == eltorito_entry.dirlevel
                        and len(eltorito_entry.bootfile_parts) > 1
                        and eltorito_entry.bootfile_parts[0] == basename
                    ):
                        eltorito_entry.bootfile_iso_path += (
                            "/" + mangle_dir_for_iso9660(basename, args.iso_level)
                        )
                        eltorito_entry.bootfile_parts.pop(0)
                        eltorito_entry.dirlevel = parent

                for f in os.listdir(localpath):
                    fullpath = os.path.join(localpath, f)

                    if (
                        match_entry_to_list(exclude_patterns, f)
                        or eltorito_duplicate_check == f
                    ):
                        print("Excluded by match: %s" % (fullpath), file=logfp)
                        continue

                    if match_entry_to_list(ignore_patterns, f):
                        print("Ignoring file %s" % (fullpath), file=logfp)
                        continue

                    if args.verbose:
                        print("Scanning %s" % (fullpath), file=logfp)

                    entries.append((fullpath, parent, True, on_eltorito_catalog_path))
            else:
                iso_path = build_iso_path(parent_level, basename, args.iso_level, False)
                if iso_path is None:
                    print(
                        "Could not find free ISO9660 name for path %s; skipping"
                        % (localpath),
                        file=logfp,
                    )

                thishash = 0
                if args.scan_for_duplicates:
                    thishash = mm3hashfromfile(localpath)

                if thishash in seen_hashes:
                    iso.add_hard_link(
                        iso_old_path=seen_hashes[thishash],
                        iso_new_path=iso_path,
                        rr_name=rr_name,
                    )
                    if joliet_path is not None:
                        iso.add_hard_link(
                            iso_old_path=seen_hashes[thishash],
                            joliet_new_path=joliet_path,
                        )
                    if udf_path is not None:
                        iso.add_hard_link(
                            iso_old_path=seen_hashes[thishash], udf_new_path=udf_path
                        )
                else:
                    iso.add_file(
                        localpath,
                        iso_path,
                        rr_name=rr_name,
                        joliet_path=joliet_path,
                        udf_path=udf_path,
                    )
                    if match_entry_to_list(hide_patterns, basename):
                        iso.rm_hard_link(iso_path=iso_path)

                    if args.joliet and match_entry_to_list(
                        hide_joliet_patterns, basename
                    ):
                        iso.rm_hard_link(joliet_path=joliet_path)

                    if args.udf and match_entry_to_list(hide_udf_patterns, basename):
                        iso.rm_hard_link(udf_path=udf_path)

                    if args.scan_for_duplicates:
                        seen_hashes[thishash] = iso_path

            if match_entry_to_list(hidden_patterns, basename):
                iso.set_hidden(iso_path)
                print("Hidden ISO9660 attribute: %s" % (localpath), file=logfp)

    # Add in El Torito if it was requested
    for entry in eltorito_entries:
        try:
            iso.add_eltorito(
                entry.bootfile_iso_path,
                bootcatfile=eltorito_catalog_path,
                bootable=entry.boot,
                boot_load_size=entry.load_size,
                boot_info_table=entry.boot_info_table,
                media_name=entry.mediatype,
                boot_load_seg=entry.load_seg,
            )
        except pycdlib.pycdlibexception.PyCdlibInvalidInput as e:
            if "Could not find path" in str(e):
                print(
                    "Uh oh, I cant find the boot image '%s' !" % (entry.bootfile_orig),
                    file=logfp,
                )
                sys.exit(255)
            else:
                raise

    class ProgressData(object):
        """
        A private class to hold onto the data from the last progress call.
        """

        __slots__ = ("last_percent", "logfp", "begun")

        def __init__(self, logfp):
            self.last_percent = ""
            self.logfp = logfp
            self.begun = time.time()

    def progress_cb(done, total, progress_data):
        """
        A private function that will be passed into the write_fp method of the
        PyCdlib object, and prints out the current progress of the mastering.

        Parameters (as prescribe by PyCdlib):
         done - The amount of data written so far
         total - The total amount of data to write
         progress_data - An object of type ProgressData to track progress
        Returns:
         Nothing.
        """
        frac = float(done) / float(total)
        percent = "%.2f%%" % (frac * 100)
        if percent != progress_data.last_percent:
            the_end = time.time()
            if frac > 0:
                the_end = progress_data.begun + (the_end - progress_data.begun) / frac
            print(
                "%s done, estimate finish %s" % (percent, time.ctime(the_end)),
                file=progress_data.logfp,
            )
            progress_data.last_percent = percent

    iso.write_fp(fp, progress_cb=progress_cb, progress_opaque=ProgressData(logfp))

    if args.print_size:
        print(
            "Total extents scheduled to be written = %d" % (len(fp.getvalue()) / 2048),
            file=logfp,
        )

    iso.close()


if __name__ == "__main__":
    main()



from distutils.core import setup
from setuptools import setup, find_packages

# Get version variables
exec(open('odarchive/_version.py').read())

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup(
    name = 'odarchive',
    version = __version__,
    description = 'Convert file systems to archiveable ISO files',
    long_description=readme + '\n\n' + history,
    author=__author__,
    author_email=__email__,
    url='https://github.com/drummonds/odarchive',
    packages=find_packages(include=['odarchive']),
    include_package_data=True,
    #install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords = ['cdrom', 'dvd', 'bdrom', 'archive', 'odarchive'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    #tests_require=test_requirements,
    #setup_requires=setup_requirements,
    #entry_points={
    #    'console_scripts': ['fab-support=fab_support.command_line:main'],
    #}
)


"""This is an database of files.

It is based on a list of entries.

It is centred on the file system.  Each file has a separate entry  It can be enhanced.
This is a temporary structure and although can be preserved by pickle is not intended to have a long duration.
So it does not need versioning.

Things it doesn't do:

- Segment the database.  See hash_db
"""
from collections import OrderedDict


def do_hash(entry):
    """Make an easy parallel task"""
    entry.calculate_file_hash()

class AbstractFileDatabaseError(Exception):
    pass


class AbstractFileDatabase:
    """This is an abstract file database which is built from a source
    However doesn't actually have the mechanics to do the building."""

    def __init__(self, source):
        super().__init__()
        self.entries = OrderedDict()  # of FileEntry
        # from source build a list of files

    def files(self):
        for entry in self.entries.values():
            yield entry

    def print_files(self):
        for entry in self.files():
            print(entry)

    def __len__(self):
        return len(self.entries)

    def get_info(self):
        """Returns summary information on a file database. Uses introspection to collect statistics"""
        result = f"Number of entries = {len(self):,}\n"
        return result



import click
from pathlib import Path

from .archive import Archiver, load_archiver_from_dill


@click.group()
def cli():
    pass


@click.command()
@click.argument("usb_path")  # , help='Path to USB drive which is to be backed up')
def init(usb_path):
    ar = Archiver()
    ar.create_file_database(Path(usb_path))
    ar.convert_to_hash_database()
    ar.hash_db.save()  # Creates catalogue.json
    ar.print_files()
    ar.save()


@click.command()
@click.argument("size")  # , help='Max size in Bytes for segment')
def segment(size):
    """Converts an archive into a segmented archive."""
    # Todo if an archive is modified eg adding new files then will need to be resegmented
    # However size parameter can't change
    ar = load_archiver_from_dill()
    ar.create_catalogue()
    ar.segment(size)


@click.command()
@click.option("--pretend", default=False, help="Won't create database if --pretend")
def write_iso(pretend):
    ar = load_archiver_from_dill()
    ar.print_files()
    ar.write_iso(pretend)
    ar.save()


@click.command()
@click.option("--pretend", default=False, help="Won't create database if --pretend")
@click.argument("usb_path")  # , help='Path to USB drive which is to be backed up')
def archive(pretend, usb_path):
    ar = Archiver()
    ar.create_file_database(Path(usb_path))
    ar.convert_to_hash_database()
    ar.hash_db.save()  # Creates catalogue.json
    ar.print_files()
    ar.write_iso(pretend)
    ar.save()



# -*- encoding: utf-8 -*-
import datetime as dt
import dateutil.parser

try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO
import logging
import os
from os import lstat
import dill

import pycdlib

from .consts import *
from .file_db import FileDatabase
from .hash_db import *
from .hash_file_entry import iso9660_dir, HashFileEntry


# import tarfile
# import tempfile
# from datetime import datetime
# from getpass import getpass
# import hashlib
# import uuid
# import socket
# import re
# import fnmatch
# import mimetypes
# import calendar
# import functools
# from contextlib import closing  # for Python2.6 compatibility
# from gzip import GzipFile
#
# import yaml
# from beefish import decrypt, encrypt_file
# import grandfatherson
# from byteformat import ByteFormatter
#
# from bakthat.backends import GlacierBackend, S3Backend, RotationConfig, SwiftBackend
# from bakthat.conf import config, events, load_config, DEFAULT_DESTINATION, DEFAULT_LOCATION, CONFIG_FILE, EXCLUDE_FILES
# from bakthat.utils import _interval_string_to_seconds
# from bakthat.models import Backups
# from bakthat.sync import BakSyncer, bakmanager_hook, bakmanager_periodic_backups
# from bakthat.plugin import setup_plugins, plugin_setup
#
# __version__ = "0.6.0"
#
#
log = logging.getLogger("odarchive")


def archive():
    pass


class odarchiveFilter(logging.Filter):

    def filter(self, rec):
        if rec.name.startswith("odarchive") or rec.name == "root":
            return True
        else:
            return rec.levelno >= logging.WARNING


#!/usr/bin/env python3
from argparse import ArgumentParser
from fnmatch import fnmatch
from os import fsdecode, fsencode, getcwd, lstat, readlink, stat_result
from pathlib import Path
import re


# fnmatch patterns, specifically:
IMPORT_FILENAME_PATTERNS = [
    DB_FILENAME,
    HASH_FILENAME,
    HASH_FILENAME + ".asc",
    "*.sha512sum",
    "*.sha512sum.asc",
    "DIGESTS",
    "DIGESTS.asc",
]
SURROGATE_ESCAPES = re.compile(r"([\udc80-\udcff])")

ADDED_COLOR = "\033[01;32m"
REMOVED_COLOR = "\033[01;34m"
MODIFIED_COLOR = "\033[01;31m"
NO_COLOR = "\033[00m"


def find_external_hash_files(path: Path):
    for dirpath_str, _, filenames in walk(str(path)):
        dirpath = Path(dirpath_str).absolute()
        for filename in filenames:
            if any(fnmatch(filename, pattern) for pattern in IMPORT_FILENAME_PATTERNS):
                yield dirpath / filename


def print_file_list(files):
    for filename in sorted(files):
        printable_filename = SURROGATE_ESCAPES.sub("\ufffd", str(filename))
        print(printable_filename)
    print()

def print_file_lists(added, removed, modified):
    if added:
        print(ADDED_COLOR + "Added files:" + NO_COLOR)
        print_file_list(added)
    if removed:
        print(REMOVED_COLOR + "Removed files:" + NO_COLOR)
        print_file_list(removed)
    if modified:
        print(MODIFIED_COLOR + "Modified files:" + NO_COLOR)
        print_file_list(modified)


def list_dirs(files):
    for filename in sorted(files):
        printable_filename = SURROGATE_ESCAPES.sub("\ufffd", str(filename))
        print(printable_filename)
    print()


def load_archiver_from_dill(filename="archiver.dill"):
    with open(filename, "rb") as f:
        # The protocol version used is detected automatically, so we do not
        # have to specify it.
        archiver = dill.load(f)
    return archiver


def load_archiver_from_json(filename):
    ar = Archiver()
    with open(filename) as json_data:
        d = json.load(json_data)
        print(d)
        print(f'guid = {d["guid"]}')
        print(f'version = {d["version"]}')
        # Create empty hash database, default is to use a file_db
        ar.hash_db = HashDatabase(None, ar.iso_path_root)
        ar.hash_db.entries = HashFileEntries.create_from_json(ar.iso_path_root, d['files'], ar.hash_db)
        """Save the current catalogue to file as a JSON file.
        It should be possible to reread this file later and recreate this record."""
        ar.hash_db.guid = uuid.UUID(d["guid"])
        if int(d['version']) != ar.hash_db.version:
            raise odarchiveError(f"Version of Catalogue ({d['version']} " +
                                 " does not match that of software ({ar.version})." +
                                 "Upgrading version not currently possible. Contact supplier")
        try:
            date_str = d['date']
            ar.archive_date = dateutil.parser.parse(date_str) # This a read only value after an archive has been
            # created
        except:
            raise odarchiveError('Date string malformed or not found in catalogue.json')
        try:
            ar.hash_db.str_segment_size = d['segment_size_str']
        except:
            ar.hash_db.str_segment_size = 'bd'
        try:
            ar.hash_db.int_segment_size = d['segment_size_int']
        except:
            ar.hash_db.int_segment_size = 25000000000
    ar.hash_db.catalogue_size = os.path.getsize(filename)
    return ar


class Archiver:
    """This holds the information on the archiving project - potentially should keep state over multiple
    invocations.  This means that you do not have to hold in memory a temporary copy of all discs but
    can do them one by one.  For a 10TB archiving to 25GB drives this is a big saving."""

    def __init__(self):
        self.iso_path_root = PurePosixPath("/DATA")

    def save(self, filename="archiver.dill"):
        with open(filename, "wb") as f:
            # Pickle the 'data' dictionary using the highest protocol available.
            dill.dump(self, f, dill.HIGHEST_PROTOCOL)

    def create_file_database(self, usb_path, job_name="new"):
        # Create database
        self.file_db = FileDatabase(usb_path)
        self.job_name = job_name
        # Check to make sure not overwriting database
        print("Initializing file database")
        self.file_db.update(
            usb_path
        )  # Scan directory to add files
        # Need to load the hash files into the Has list

    def convert_to_hash_database(self, verbose=False):
        if not self.is_locked:
            self.file_db.calculate_file_hash(verbose)
            # Create database
            self.hash_db = HashDatabase(self.file_db, self.iso_path_root)
        else:
            raise odarchiveError('Archive locked so cannot calculate hashes')

    def create_catalogue(self, verbose=False):
        """Creates a catalogue file catalogue.json on disk."""
        self.hash_db.save()  # Creates catalogue.json
        self.save()


    def write_iso(self, pretend=False, disc_num=None, job_name="new"):
        """No ISO file will be created if there are not files in it.  Eg using a disc num that is
        not being used."""
        try:
            if disc_num is None and self.hash_db.last_disc_number is not None:
                raise odarchiveError("disc_num is None but archive has been segmented")
            elif self.hash_db.last_disc_number is None and disc_num is not None:
                raise odarchiveError(
                    f"disc_num is {disc_num} but archive has not been segmented"
                )
            else:
                # Create ISO file
                iso = pycdlib.PyCdlib()
                if disc_num is None:  # Assuming standard is one based.
                    seqnum = 0
                else:
                    seqnum = disc_num
                if self.last_disc_num is None:  # Has not been segmented
                    set_size = 1
                else:
                    set_size = self.last_disc_num
                print(f'Disc num = |{disc_num}|')
                iso.new(
                    interchange_level=3,
                    udf="2.60",
                    app_ident_str="odarchive (C) 2018 drummonds.net",
                    abstract_file="README.MKD",
                    set_size=set_size,  # This is the size of the set of discs ie the number of disck
                    # eg 2 of set_size
                    seqnum=seqnum,
                )
        except AttributeError:
            raise odarchiveError(
                f"Probably have not yet created hash db"
            )

        # The readme file is created fresh for each disc created.  It should consist of specific information about
        # this disc and also information about the archive process
        readme = f"""# Archive File created by www.drummonds.net
This archive was created {dt.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}

The data for this archive is stored in the directory /DATA.
There is a catalogue of this archive stored in catalouge.json.  This catalogue has a list of all the files
archived in this run and on which disc they are stored.  The same catalogue is written to each disc in the 
archive series."""
        readme_bytes = readme.encode("utf-8")
        iso.add_fp(
            BytesIO(readme_bytes),
            len(readme_bytes),
            "/README.MKD;1",
            udf_path="/README.MKD",
        )
        iso.add_file(
            str("catalogue.json"),
            "/CATALOGUE.JSON;1",
            udf_path="/catalogue.json"  # Same catalogue for each disc
            # So that you can go to single disc and then find where to go next - which disc to read rather than
            # having to read all the files.
        )
        # Todo add_file which says which disc this is in which catalogue

        # iso.add_directory(str(self.iso_path_root), udf_path=str(self.iso_path_root))
        # iso.add_directory(
        #     str(self.iso_path_root) + "/TESTDIR",
        #     udf_path=str(self.iso_path_root) + "/testDir",
        # )
        dir_count = 0
        for this_dir in self.hash_db.entries.dir_entries(disc_num=disc_num):
            # Todo add Bridge format and iso9660
            # iso.add_directory(iso9660_dir(this_dir), udf_path=this_dir)
            # After 10^8 directories (which breaks a standard ISO 9660 the formatting will vary
            if this_dir == "/DATA":
                iso.add_directory(
                    "/DATA", udf_path="/DATA"
                )  # Add root data directory to both ISO and UDF
            else:
                iso.add_directory(
                    f"/DATA/{dir_count:08}", udf_path=this_dir
                )  # Note can't use "/" as ISO 9660 root as we are adding
                # a directory and this would only be the root
            dir_count += 1
        any_files = False
        file_count = 0
        for this_file in self.hash_db.files(disc_num=disc_num):
            # Todo add Bridge format and iso9660
            # iso.add_file(
            #     str(this_file.filename),
            #     this_file.iso9660_path,
            #     udf_path=str(this_file.udf_absolute_path),
            # )
            iso.add_file(
                str(this_file.file_system_path),
                f"/DATA/{file_count:08}",  # All data files in same directory and anonymise names :(
                udf_path=str(this_file.udf_absolute_path),
            )
            any_files = True
            file_count += 1
        if (
            not pretend and any_files
        ):  # Will not write out a cataloge with no files in it
            if disc_num is None:
                filename = f"{job_name}.iso"
            else:
                filename = f"{job_name}_{disc_num:04}.iso"
            # If file exists then iso.write will just overwrite part of it so need to delete it first.
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass
            iso.write(filename)
            iso.close()

    @property
    def is_locked(self):
        """Once you have successfully written the first ISO then you should lock the archive.
        This should prevent operations such as segmenting if the archive is locked."""
        try:
            return self.locked
        except AttributeError:  # NO hash db so not segmented
            return False

    def segment(self, size):
        """Segment an archive.  This is mainly"""
        if not self.is_locked:
            catalogue_size = (
                (2048 + lstat(str(str("catalogue.json"))).st_size) // 2048
            ) * 2048  # Account for sector size
            self.hash_db.segment(size, catalogue_size)
        else:
            raise odarchiveError('Archive is locked so cannot resegment')

    @property
    def is_segmented(self):
        try:
            return self.hash_db.is_segmented
        except AttributeError:  # NO hash db so not segmented
            return False

    def get_info(self):
        """Returns summary information on archive"""
        try:
            result = self.hash_db.get_info()
        except:
            result = self.file_db.get_info()
        return result

    def print_files(self):
        self.file_db.print_files()

    @property
    def last_disc_num(self):
        return self.hash_db.last_disc_number

    @property
    def guid(self):
        try:
            return self.hash_db.guid
        except AttributeError:
            return None

    def get_disc_info(self, disc_num):
        """Returns summary information on a single disc"""
        return self.hash_db.get_info(disc_num)

    def get_all_disc_info(self):
        """Returns summary information on all discs"""
        result = ''
        for i in range(self.last_disc_num):
            result += self.get_disc_info(i)
        return result
